"""
Data structures for intent recognition.
"""
import dataclasses
import datetime
import typing
from collections.abc import Mapping
from dataclasses import dataclass, field
from enum import Enum
from numbers import Number

from . import utils


@dataclass
class Entity:
    """Named entity from intent."""

    entity: str
    value: typing.Any
    raw_value: str = ""
    source: str = ""
    start: int = 0
    raw_start: int = 0
    end: int = 0
    raw_end: int = 0
    tokens: typing.List[typing.Any] = field(default_factory=list)
    raw_tokens: typing.List[str] = field(default_factory=list)

    @property
    def value_dict(self):
        """Get dictionary representation of value."""
        if isinstance(self.value, Mapping):
            return self.value

        kind = "Unknown"

        if isinstance(self.value, Number):
            kind = "Number"
        elif isinstance(self.value, datetime.date):
            kind = "Date"
        elif isinstance(self.value, datetime.time):
            kind = "Time"
        elif isinstance(self.value, datetime.datetime):
            kind = "Datetime"
        elif isinstance(self.value, datetime.timedelta):
            kind = "Duration"

        return {"kind": kind, "value": self.value}

    @classmethod
    def from_dict(cls, entity_dict: typing.Dict[str, typing.Any]) -> "Entity":
        """Create Entity from dictionary."""
        return Entity(**utils.only_fields(cls, entity_dict))


@dataclass
class Intent:
    """Named intention with entities and slots."""

    name: str
    confidence: float = 0

    @classmethod
    def from_dict(cls, intent_dict: typing.Dict[str, typing.Any]) -> "Intent":
        """Create Intent from dictionary."""
        return Intent(**utils.only_fields(cls, intent_dict))


@dataclass
class TagInfo:
    """Information used to process FST tags."""

    tag: str
    start_index: int = 0
    raw_start_index: int = 0
    symbols: typing.List[str] = field(default_factory=list)
    raw_symbols: typing.List[str] = field(default_factory=list)

    @classmethod
    def from_dict(cls, tag_dict: typing.Dict[str, typing.Any]) -> "TagInfo":
        """Create TagInfo from dictionary."""
        return TagInfo(**utils.only_fields(cls, tag_dict))


class RecognitionResult(str, Enum):
    """Result of a recognition."""

    SUCCESS = "success"
    FAILURE = "failure"


@dataclass
class Recognition:
    """Output of intent recognition."""

    intent: typing.Optional[Intent] = None
    entities: typing.List[Entity] = field(default_factory=list)
    text: str = ""
    raw_text: str = ""
    recognize_seconds: float = 0
    tokens: typing.List[typing.Any] = field(default_factory=list)
    raw_tokens: typing.List[str] = field(default_factory=list)

    # Transcription details
    wav_seconds: typing.Optional[float] = None
    transcribe_seconds: typing.Optional[float] = None
    speech_confidence: typing.Optional[float] = None
    wav_name: typing.Optional[str] = None

    def asdict(self) -> typing.Dict[str, typing.Any]:
        """Convert to dictionary."""
        return dataclasses.asdict(self)

    @classmethod
    def empty(cls) -> "Recognition":
        """Return an empty recognition."""
        return Recognition(intent=Intent(name=""))

    @classmethod
    def from_dict(cls, recognition_dict: typing.Dict[str, typing.Any]) -> "Recognition":
        """Create Recognition from dictionary."""

        intent_dict = recognition_dict.pop("intent", None)
        entity_dicts = recognition_dict.pop("entities", None)
        slots_dict = recognition_dict.pop("slots", None)
        recognition = Recognition(**utils.only_fields(cls, recognition_dict))

        if intent_dict:
            recognition.intent = Intent.from_dict(intent_dict)

        if entity_dicts:
            recognition.entities = [Entity.from_dict(e) for e in entity_dicts]

        if slots_dict:
            recognition.entities = [
                Entity(entity=key, value=value) for key, value in slots_dict.items()
            ]

        return recognition



"""Types and constants."""
import typing

from .jsgf import Expression, Rule, Sentence

IntentsType = typing.Dict[str, typing.List[typing.Union[Sentence, Rule]]]
SentencesType = typing.Dict[str, typing.List[Sentence]]
ReplacementsType = typing.Dict[str, typing.List[Expression]]



"""Number range expansion."""
import logging
import re
import typing

from num2words import num2words

from .jsgf import Expression, Sequence, SequenceType, SlotReference, Word

_LOGGER = logging.getLogger(__name__)

# 0..100, -100..100
NUMBER_RANGE_PATTERN = re.compile(r"^(-?[0-9]+)\.\.(-?[0-9]+)(,[0-9]+)?$")
NUMBER_PATTERN = re.compile(r"^(-?[0-9]+)$")

# -----------------------------------------------------------------------------


def number_to_words(
    number: int, language: typing.Optional[str] = None
) -> typing.List[str]:
    """Convert number to list of words (75 -> seventy five)"""
    language = language or "en"
    number_text = num2words(number, lang=language).replace("-", " ").strip()
    return number_text.split()


def replace_numbers(
    words: typing.Iterable[str], language: typing.Optional[str] = None
) -> typing.Iterable[str]:
    """Replace numbers with words in a sentence (75 hats -> seventy five hats)"""
    language = language or "en"
    for word in words:
        if NUMBER_PATTERN.match(word):
            n = int(word)
            for number_word in number_to_words(n, language=language):
                yield number_word
        else:
            yield word


def number_range_transform(word: Expression, slot_name="rhasspy/number"):
    """Automatically transform number ranges to slot reference (e.g., 0..100)"""
    if not isinstance(word, Word):
        # Skip anything besides words
        return

    match = NUMBER_RANGE_PATTERN.match(word.text)

    if not match:
        return

    try:
        lower_bound = int(match.group(1))
        upper_bound = int(match.group(2))
        step = 1

        if len(match.groups()) > 3:
            # Exclude ,
            step = int(match.group(3)[1:])

        # Transform to $rhasspy/number
        return SlotReference(
            text=word.text,
            slot_name=f"{slot_name},{lower_bound},{upper_bound},{step}",
            converters=["int"],
        )
    except ValueError:
        # Not a number
        pass
    except Exception:
        _LOGGER.exception("number_range_transform")


def number_transform(word: Expression, language: typing.Optional[str] = None):
    """Automatically transform numbers to words (e.g., 75)"""
    if not isinstance(word, Word):
        # Skip anything besides words
        return

    match = NUMBER_PATTERN.match(word.text)

    if not match:
        return

    try:
        n = int(match.group(1))

        # 75 -> (seventy five):75!int
        number_words = number_to_words(n, language=language)

        if len(number_words) == 1:
            # Easy case, single word
            word.text = number_words[0]
            word.substitution = [str(n)]
            word.converters = ["int"]
            return word

        # Hard case, split into multiple Words
        number_text = " ".join(number_words)
        return Sequence(
            text=number_text,
            type=SequenceType.GROUP,
            substitution=[str(n)],
            converters=["int"],
            items=[Word(w) for w in number_words],
        )
    except ValueError:
        # Not a number
        pass
    except Exception:
        _LOGGER.exception("number_transform")



"""Methods for evaluating recognition results."""
import logging
import typing
from dataclasses import dataclass, field

from .intent import Recognition

_LOGGER = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


@dataclass
class WordError:
    """Detailed differences between reference and hypothesis strings."""

    reference: typing.List[str] = field(default_factory=list)
    hypothesis: typing.List[str] = field(default_factory=list)
    differences: typing.List[str] = field(default_factory=list)
    words: int = 0
    errors: int = 0
    matches: int = 0
    substitutions: int = 0
    deletions: int = 0
    insertions: int = 0
    error_rate: float = 0.0


@dataclass
class TestReportItem(Recognition):
    """Extended actual recognition result from TestReport."""

    expected_intent_name: str = ""
    wrong_entities: typing.List[typing.Tuple[str, str]] = field(default_factory=list)
    missing_entities: typing.List[typing.Tuple[str, str]] = field(default_factory=list)
    word_error: typing.Optional[WordError] = None


@dataclass
class TestReport:
    """Result of evaluate_intents."""

    expected: typing.Dict[str, Recognition] = field(default_factory=dict)
    actual: typing.Dict[str, TestReportItem] = field(default_factory=dict)

    # ----------
    # Statistics
    # ----------

    # Total number of WAV files
    num_wavs: int = 0

    # Number of words in all transcriptions (as counted by word_align.pl)
    num_words: int = 0

    # Total number of intents that were attempted
    num_intents: int = 0

    # Number of entity/value pairs all intents
    num_entities: int = 0

    # Number transcriptions that match *exactly*
    correct_transcriptions: int = 0

    # Number of recognized intents that match expectations
    correct_intent_names: int = 0

    # Number of correct words in all transcriptions (as computed by word_align.pl)
    correct_words: int = 0

    # Number of entity/value pairs that match *exactly* in all recognized intents
    correct_entities: int = 0

    # Number of intents where name and entities match exactly
    correct_intent_and_entities: int = 0

    transcription_accuracy: typing.Optional[float] = None

    intent_accuracy: typing.Optional[float] = None

    entity_accuracy: typing.Optional[float] = None

    intent_entity_accuracy: typing.Optional[float] = None

    # Average wav seconds / transcribe seconds
    average_transcription_speedup: typing.Optional[float] = None


# -----------------------------------------------------------------------------


def evaluate_intents(
    expected: typing.Dict[str, Recognition], actual: typing.Dict[str, Recognition]
) -> TestReport:
    """Generate report of comparison between expected and actual recognition results."""
    # Actual intents and extra info about missing entities, etc.
    report = TestReport(expected=dict(expected))

    # Real time vs transcription time
    speedups = []

    # Compute statistics
    for wav_name, actual_intent in actual.items():
        # pylint: disable=E1137
        report.actual[wav_name] = TestReportItem(**actual_intent.asdict())

        # Get corresponding expected intent
        expected_intent = expected[wav_name]

        # Compute real-time speed-up
        wav_seconds = actual_intent.wav_seconds or 0.0
        transcribe_seconds = actual_intent.transcribe_seconds or 0.0
        if (transcribe_seconds > 0) and (wav_seconds > 0):
            speedups.append(wav_seconds / transcribe_seconds)

        # Check transcriptions
        actual_text = actual_intent.raw_text or actual_intent.text
        expected_text = expected_intent.raw_text or expected_intent.text

        if expected_text == actual_text:
            report.correct_transcriptions += 1

        # Check intents
        if expected_intent.intent is not None:
            report.num_intents += 1
            if actual_intent.intent is None:
                intents_match = False
            else:
                intents_match = expected_intent.intent.name == actual_intent.intent.name

            # Count entities
            expected_entities: typing.List[typing.Tuple[str, str]] = []
            num_expected_entities = 0
            for entity in expected_intent.entities:
                report.num_entities += 1
                num_expected_entities += 1
                entity_tuple = (entity.entity, entity.value)
                expected_entities.append(entity_tuple)

            # Verify actual entities.
            # Only check entities if intent was correct.
            wrong_entities: typing.List[typing.Tuple[str, str]] = []
            missing_entities: typing.List[typing.Tuple[str, str]] = []

            if intents_match:
                report.correct_intent_names += 1
                num_actual_entities = 0
                for entity in actual_intent.entities:
                    num_actual_entities += 1
                    entity_tuple = (entity.entity, entity.value)

                    if entity_tuple in expected_entities:
                        report.correct_entities += 1
                        expected_entities.remove(entity_tuple)
                    else:
                        wrong_entities.append(entity_tuple)

                # Anything left is missing
                missing_entities = expected_entities

                # Check if entities matched *exactly*
                if (not expected_entities) and (
                    num_actual_entities == num_expected_entities
                ):
                    report.correct_intent_and_entities += 1

            # pylint: disable=E1136
            report.actual[wav_name].expected_intent_name = expected_intent.intent.name

            # pylint: disable=E1136
            report.actual[wav_name].wrong_entities = wrong_entities

            # pylint: disable=E1136
            report.actual[wav_name].missing_entities = missing_entities

        # Compute word error
        if expected_text:
            word_error = get_word_error(expected_text.split(), actual_text.split())
            report.num_words += word_error.words
            report.correct_words += word_error.matches

            # pylint: disable=E1136
            report.actual[wav_name].word_error = word_error

        report.num_wavs += 1

    # ---------------------------------------------------------------------

    if report.num_wavs < 1:
        _LOGGER.warning("No WAV files found")

    # Compute transcription speedup
    if speedups:
        report.average_transcription_speedup = sum(speedups) / len(speedups)

    # Summarize results
    if report.num_words > 0:
        report.transcription_accuracy = report.correct_words / report.num_words

    if report.num_intents > 0:
        report.intent_accuracy = report.correct_intent_names / report.num_intents

    if report.num_entities > 0:
        report.entity_accuracy = report.correct_entities / report.num_entities

    if report.num_intents > 0:
        report.intent_entity_accuracy = (
            report.correct_intent_and_entities / report.num_intents
        )

    return report


# -----------------------------------------------------------------------------

# Reference: https://github.com/jtsi/asr-wer


def get_word_error(
    reference: typing.List[str], hypothesis: typing.List[str]
) -> WordError:
    """Computes insertions/deletions/substituions and word error rate for two sequences of words."""
    if not reference:
        raise ValueError("Reference cannot be empty")

    reference = [w.lower() for w in reference]
    hypothesis = [w.lower() for w in hypothesis]

    if not hypothesis:
        # All error
        return WordError(
            reference=reference,
            hypothesis=hypothesis,
            differences=[w.upper() for w in hypothesis],
            insertions=len(hypothesis),
            error_rate=1,
        )

    # Initialize the matrix/table and set the first row and column equal to
    # 1, 2, 3, ...
    # Each column represent a single token in the reference string a
    # Each row represent a single token in the reference string b
    #
    rows = len(hypothesis) + 1
    cols = len(reference) + 1
    m = [[0] * cols for _ in range(rows)]

    for col in range(cols):
        for row in range(rows):
            if row == 0:
                m[0][col] = col
            elif col == 0:
                m[row][0] = row

    # Now loop over remaining cell (from the second row and column onwards)
    # The value of each selected cell is:
    #
    #   if token represented by row == token represented by column:
    #       value of the top-left diagonal cell
    #   else:
    #       calculate 3 values:
    #            * top-left diagonal cell + 1 (which represents substitution)
    #            * left cell + 1 (representing deleting)
    #            * top cell + 1 (representing insertion)
    #       value of the smallest of the three
    #
    for row in range(1, rows):
        for col in range(1, cols):
            if reference[col - 1] == hypothesis[row - 1]:
                m[row][col] = m[row - 1][col - 1]
            else:
                substitution = m[row - 1][col - 1] + 1
                insertion = m[row][col - 1] + 1
                deletion = m[row - 1][col] + 1
                m[row][col] = min(substitution, insertion, deletion)

    # and the minimum-edit distance is simply the value of the down-right most
    # cell
    errors = m[rows - 1][cols - 1]

    # Compute statistics and differences
    ref_index = len(reference)
    hyp_index = len(hypothesis)
    differences = []
    matches = 0
    substitutions = 0
    insertions = 0
    deletions = 0
    while True:
        if (ref_index <= 0) or (hyp_index <= 0):
            break

        if reference[ref_index - 1] == hypothesis[hyp_index - 1]:
            # Match
            matches += 1
            same_word = hypothesis[hyp_index - 1]
            differences.append(same_word)
            ref_index -= 1
            hyp_index -= 1
        elif m[hyp_index][ref_index] == (m[hyp_index - 1][ref_index - 1] + 1):
            # Substitution
            substitutions += 1
            ref_word = reference[ref_index - 1]
            hyp_word = hypothesis[hyp_index - 1]
            differences.append(f"{ref_word}:{hyp_word}")
            ref_index -= 1
            hyp_index -= 1
        elif m[hyp_index][ref_index] == (m[hyp_index][ref_index - 1] + 1):
            # Deletion
            deletions += 1
            ref_word = reference[ref_index - 1]
            differences.append(f"-{ref_word}")
            ref_index -= 1
        elif m[hyp_index][ref_index] == (m[hyp_index - 1][ref_index] + 1):
            # Insertion
            insertions += 1
            hyp_word = hypothesis[hyp_index - 1]
            differences.append(f"+{hyp_word}")
            hyp_index -= 1
        else:
            _LOGGER.warning("get_word_error: malformed matrix")
            break

    # error = (S + D + I) / N
    error_rate = (substitutions + deletions + insertions) / len(reference)

    return WordError(
        reference=reference,
        hypothesis=hypothesis,
        differences=list(reversed(differences)),
        words=len(reference),
        matches=matches,
        substitutions=substitutions,
        insertions=insertions,
        deletions=deletions,
        errors=errors,
        error_rate=error_rate,
    )



"""Test cases for evaluation."""
import unittest

from rhasspynlu.evaluate import evaluate_intents, get_word_error
from rhasspynlu.intent import Entity, Intent, Recognition


class WordErrorEvaluationTests(unittest.TestCase):
    """Test cases for word error."""

    def test_wer_perfect(self):
        """Test word error rate calculation with perfect match."""
        reference = "this is a test".split()
        hypothesis = reference

        result = get_word_error(reference, hypothesis)
        self.assertEqual(4, result.matches)
        self.assertEqual(0, result.errors)
        self.assertEqual(0, result.substitutions)
        self.assertEqual(0, result.insertions)
        self.assertEqual(0, result.deletions)
        self.assertEqual(0, result.error_rate)

        self.assertEqual(reference, result.reference)
        self.assertEqual(hypothesis, result.hypothesis)
        self.assertEqual(reference, result.differences)

    def test_wer_substitution(self):
        """Test word error rate calculation with a substitution."""
        reference = "this is a test".split()
        hypothesis = "this is bad test".split()

        result = get_word_error(reference, hypothesis)
        self.assertEqual(1, result.errors)
        self.assertEqual(3, result.matches)
        self.assertEqual(1, result.substitutions)
        self.assertEqual(0, result.insertions)
        self.assertEqual(0, result.deletions)
        self.assertEqual(0.25, result.error_rate)

        self.assertEqual(reference, result.reference)
        self.assertEqual(hypothesis, result.hypothesis)
        self.assertEqual(["this", "is", "a:bad", "test"], result.differences)

    def test_wer_multiple(self):
        """Test word error rate calculation with a multiple insertions/deletions/substitutions."""
        reference = "this is a test".split()
        hypothesis = "this bad test test opps".split()

        result = get_word_error(reference, hypothesis)
        self.assertEqual(3, result.errors)
        self.assertEqual(2, result.matches)
        self.assertEqual(2, result.substitutions)
        self.assertEqual(1, result.insertions)
        self.assertEqual(0, result.deletions)
        self.assertEqual(0.75, result.error_rate)

        self.assertEqual(reference, result.reference)
        self.assertEqual(hypothesis, result.hypothesis)
        self.assertEqual(
            ["this", "is:bad", "a:test", "test", "+opps"], result.differences
        )


# -----------------------------------------------------------------------------


class IntentEvaluationTests(unittest.TestCase):
    """Test cases for intent evaluation."""

    def test_evaluate_perfect(self):
        """Test intent evaluation with a perfect match."""
        expected = {
            "test1": Recognition(
                intent=Intent(name="TestIntent"),
                entities=[Entity(entity="testEntity", value="testValue")],
                text="this is a test",
            )
        }

        actual = expected
        report = evaluate_intents(expected, actual)

        self.assertEqual(1, report.num_wavs)
        self.assertEqual(1, report.num_intents)
        self.assertEqual(1, report.num_entities)

        self.assertEqual(1, report.correct_intent_names)
        self.assertEqual(1, report.correct_entities)
        self.assertEqual(1, report.correct_intent_and_entities)


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    unittest.main()



"""Utilities for ARPA language models."""
import logging
import re
import shutil
import subprocess
import tempfile
import typing
from pathlib import Path

import networkx as nx

from .jsgf_graph import graph_to_fst
from .ngram import get_intent_ngram_counts

_LOGGER = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


def graph_to_arpa(
    graph: nx.DiGraph,
    arpa_path: typing.Union[str, Path],
    vocab_path: typing.Optional[typing.Union[str, Path]] = None,
    intent_filter: typing.Optional[typing.Callable[[str], bool]] = None,
    **fst_to_arpa_args,
):
    """Convert intent graph to ARPA language model using opengrm."""
    with tempfile.TemporaryDirectory() as temp_dir_str:
        temp_dir = Path(temp_dir_str)
        fst_text_path = temp_dir / "graph.fst.txt"
        isymbols_path = temp_dir / "isymbols.txt"
        osymbols_path = temp_dir / "osymbols.txt"

        # Graph -> binary FST
        graph_to_fst(graph, intent_filter=intent_filter).write_fst(
            fst_text_path, isymbols_path, osymbols_path
        )

        if vocab_path:
            # Extract vocabulary
            with open(vocab_path, "w") as vocab_file:
                with open(isymbols_path, "r") as isymbols_file:
                    for line in isymbols_file:
                        line = line.strip()
                        if line:
                            # symbol N
                            isymbol = line[: line.rfind(" ")]
                            if isymbol and (isymbol[0] not in ["_", "<"]):
                                print(isymbol, file=vocab_file)

                _LOGGER.debug("Wrote vocabulary to %s", vocab_path)

        # Convert to ARPA
        fst_to_arpa(
            fst_text_path, isymbols_path, osymbols_path, arpa_path, **fst_to_arpa_args
        )


def fst_to_arpa(
    fst_text_path: typing.Union[str, Path],
    isymbols_path: typing.Union[str, Path],
    osymbols_path: typing.Union[str, Path],
    arpa_path: typing.Union[str, Path],
    **kwargs,
):
    """Convert text FST to ARPA language model using opengrm."""
    for task in fst_to_arpa_tasks(
        fst_text_path, isymbols_path, osymbols_path, arpa_path, **kwargs
    ):
        run_task(task)


def graph_to_arpa_small(
    graph: nx.DiGraph,
    arpa_path: typing.Union[str, Path],
    vocab_path: typing.Optional[typing.Union[str, Path]] = None,
    dictionary_word_transform: typing.Optional[typing.Callable[[str], str]] = None,
    balance_counts: bool = True,
    estimate_ngram: typing.Optional[typing.Union[str, Path]] = None,
):
    """Convert intent graph to ARPA language model using MITLM. Works better for small graphs."""
    estimate_ngram = estimate_ngram or shutil.which("estimate-ngram")
    assert estimate_ngram, "Missing estimate-ngram in PATH"

    # Generate counts
    _LOGGER.debug("Generating ngram counts")
    intent_counts = get_intent_ngram_counts(graph, balance_counts=balance_counts)

    # Create ngram counts file
    with tempfile.NamedTemporaryFile(mode="w+") as count_file:
        for intent_name in intent_counts:
            for ngram, count in intent_counts[intent_name].items():
                if dictionary_word_transform:
                    ngram = [dictionary_word_transform(w) for w in ngram]

                # word [word] ... <TAB> count
                print(*ngram, file=count_file, end="")
                print("\t", count, file=count_file)

        count_file.seek(0)
        with tempfile.NamedTemporaryFile(mode="w+") as vocab_file:
            ngram_command = [
                str(estimate_ngram),
                "-order",
                "3",
                "-counts",
                count_file.name,
                "-write-lm",
                str(arpa_path),
                "-write-vocab",
                vocab_file.name,
            ]

            _LOGGER.debug(ngram_command)
            subprocess.check_call(ngram_command)

            if vocab_path:
                # Copy over real file
                vocab_file.seek(0)
                with open(vocab_path, "w") as real_vocab_file:
                    for line in vocab_file:
                        line = line.strip()
                        if line and (line[0] not in ["_", "<"]):
                            print(line, file=real_vocab_file)


# -----------------------------------------------------------------------------


def arpa_to_fst(arpa_path: typing.Union[str, Path], fst_path: typing.Union[str, Path]):
    """Convert ARPA language model to FST. Typically for language model mixing."""
    run_task(arpa_to_fst_task(arpa_path, fst_path))


def arpa_to_fst_task(
    arpa_path: typing.Union[str, Path], fst_path: typing.Union[str, Path]
) -> typing.Dict[str, typing.Any]:
    """Generate doit compatible task for ARPA to FST conversion."""
    return {
        "name": "base_lm_to_fst",
        "file_dep": [arpa_path],
        "targets": [fst_path],
        "actions": ["ngramread --ARPA %(dependencies)s %(targets)s"],
    }


# -----------------------------------------------------------------------------


def fst_to_arpa_tasks(
    fst_text_path: typing.Union[str, Path],
    isymbols_path: typing.Union[str, Path],
    osymbols_path: typing.Union[str, Path],
    arpa_path: typing.Union[str, Path],
    fst_path: typing.Optional[typing.Union[str, Path]] = None,
    counts_path: typing.Optional[typing.Union[str, Path]] = None,
    model_path: typing.Optional[typing.Union[str, Path]] = None,
    base_fst_weight: typing.Optional[
        typing.Tuple[typing.Union[str, Path], float]
    ] = None,
    merge_path: typing.Optional[typing.Union[str, Path]] = None,
) -> typing.Iterable[typing.Dict[str, typing.Any]]:
    """Generate doit compatible tasks for FST to ARPA conversion."""
    # Text -> FST
    fst_text_path = Path(fst_text_path)
    fst_path = Path(fst_path or fst_text_path.with_suffix(".fst"))

    yield {
        "name": "compile_fst",
        "file_dep": [fst_text_path, isymbols_path, osymbols_path],
        "targets": [fst_path],
        "actions": [
            "fstcompile "
            "--keep_isymbols --keep_osymbols "
            f"--isymbols={isymbols_path} --osymbols={osymbols_path} "
            f"{fst_text_path} %(targets)s"
        ],
    }

    # FST -> n-gram counts
    counts_path = counts_path or Path(str(fst_path) + ".counts")
    yield {
        "name": "intent_counts",
        "file_dep": [fst_path],
        "targets": [counts_path],
        "actions": ["ngramcount %(dependencies)s %(targets)s"],
    }

    # n-gram counts -> model
    model_path = model_path or Path(str(fst_path) + ".model")
    yield {
        "name": "intent_model",
        "file_dep": [counts_path],
        "targets": [model_path],
        "actions": ["ngrammake --method=witten_bell %(dependencies)s %(targets)s"],
    }

    if base_fst_weight:
        # Mixed language modeling
        base_path, base_weight = base_fst_weight
        if base_weight > 0:
            merge_path = merge_path or Path(str(fst_path) + ".merge")

            # merge
            yield {
                "name": "lm_merge",
                "file_dep": [base_path, model_path],
                "targets": [merge_path],
                "actions": [
                    "ngrammerge "
                    f"--alpha={base_weight} "
                    "%(dependencies)s %(targets)s"
                ],
            }

            # Use merged model instead
            model_path = merge_path

    # model -> ARPA
    yield {
        "name": "intent_arpa",
        "file_dep": [model_path],
        "targets": [arpa_path],
        "actions": ["ngramprint --ARPA %(dependencies)s > %(targets)s"],
    }


# -----------------------------------------------------------------------------


def mix_language_models(
    small_arpa_path: typing.Union[str, Path],
    large_fst_path: typing.Union[str, Path],
    mix_weight: float,
    mixed_arpa_path: typing.Union[str, Path],
    small_fst_path: typing.Optional[typing.Union[str, Path]] = None,
    mixed_fst_path: typing.Optional[typing.Union[str, Path]] = None,
):
    """Mix a large pre-built FST language model with a small ARPA model."""
    with tempfile.NamedTemporaryFile(suffix=".fst", mode="w+") as small_fst:
        # Convert small ARPA to FST
        if not small_fst_path:
            small_fst_path = small_fst.name

        small_command = [
            "ngramread",
            "--ARPA",
            str(small_arpa_path),
            str(small_fst_path),
        ]
        _LOGGER.debug(small_command)
        subprocess.check_call(small_command)

        with tempfile.NamedTemporaryFile(suffix=".fst", mode="w+") as mixed_fst:
            # Merge ngram FSTs
            small_fst.seek(0)

            if not mixed_fst_path:
                mixed_fst_path = mixed_fst.name

            merge_command = [
                "ngrammerge",
                f"--alpha={mix_weight}",
                str(large_fst_path),
                str(small_fst_path),
                str(mixed_fst_path),
            ]
            _LOGGER.debug(merge_command)
            subprocess.check_call(merge_command)

            # Write to final ARPA
            mixed_fst.seek(0)
            print_command = [
                "ngramprint",
                "--ARPA",
                str(mixed_fst_path),
                str(mixed_arpa_path),
            ]
            _LOGGER.debug(print_command)
            subprocess.check_call(print_command)


# -----------------------------------------------------------------------------


def get_perplexity(
    text: str, language_model_fst: typing.Union[str, Path], debug: bool = False
) -> typing.Optional[float]:
    """Compute perplexity of text with ngramperplexity."""

    with tempfile.TemporaryDirectory() as temp_dir_name:
        temp_path = Path(temp_dir_name)
        text_path = temp_path / "sentence.txt"
        text_path.write_text(text)

        symbols_path = temp_path / "sentence.syms"
        symbols_command = ["ngramsymbols", str(text_path), str(symbols_path)]
        _LOGGER.debug(symbols_command)
        subprocess.check_call(symbols_command)

        far_path = temp_path / "sentence.far"
        compile_command = [
            "farcompilestrings",
            f"-symbols={symbols_path}",
            "-keep_symbols=1",
            str(text_path),
            str(far_path),
        ]
        _LOGGER.debug(compile_command)
        subprocess.check_call(compile_command)

        verbosity = 1 if debug else 0
        perplexity_command = [
            "ngramperplexity",
            f"--v={verbosity}",
            str(language_model_fst),
            str(far_path),
        ]
        _LOGGER.debug(perplexity_command)
        output = subprocess.check_output(perplexity_command).decode()

        _LOGGER.debug(output)

        last_line = output.strip().splitlines()[-1]
        match = re.match(r"^.*perplexity\s*=\s*(.+)$", last_line)

        if match:
            perplexity = float(match.group(1))
            return perplexity

        return None


# -----------------------------------------------------------------------------


def run_task(task: typing.Dict[str, typing.Any]):
    """Execute a doit compatible task."""
    name = task.get("name", "task")
    for action in task["actions"]:
        file_dep = " ".join(f'"{d}"' for d in task.get("file_dep", []))
        targets = " ".join(f'"{t}"' for t in task.get("targets", []))
        command = action % {"dependencies": file_dep, "targets": targets}
        _LOGGER.debug("%s: %s", name, command)
        subprocess.check_call(command, shell=True)



"""Grapheme to phoneme functions for word pronunciations."""
import logging
import re
import shutil
import subprocess
import tempfile
import typing
from pathlib import Path

PronunciationsType = typing.Dict[str, typing.List[typing.List[str]]]

_LOGGER = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class MissingWordPronunciationsException(Exception):
    """Raised when missing word pronunciations and no g2p model."""

    def __init__(self, words: typing.List[str]):
        super().__init__(self)
        self.words = words

    def __str__(self):
        return f"Missing pronunciations for: {self.words}"


# -----------------------------------------------------------------------------


def read_pronunciations(
    dict_file: typing.Iterable[str],
    word_dict: typing.Optional[PronunciationsType] = None,
) -> PronunciationsType:
    """Loads a CMU-like pronunciation dictionary, optionally into an existing dictionary."""
    if word_dict is None:
        word_dict = {}

    for i, line in enumerate(dict_file):
        line = line.strip()
        if not line:
            continue

        try:
            # Use explicit whitespace (avoid 0xA0)
            word, *pronounce = re.split(r"[ \t]+", line)

            word = word.split("(")[0]

            if word in word_dict:
                word_dict[word].append(pronounce)
            else:
                word_dict[word] = [pronounce]
        except Exception as e:
            _LOGGER.warning("read_pronunciations: %s (line %s)", e, i + 1)

    return word_dict


def write_pronunciations(
    vocabulary: typing.Set[str],
    pronunciations: PronunciationsType,
    dictionary: typing.Union[str, Path],
    g2p_model: typing.Optional[typing.Union[str, Path]] = None,
    g2p_word_transform: typing.Optional[typing.Callable[[str], str]] = None,
    phonetisaurus_apply: typing.Optional[typing.Union[str, Path]] = None,
    missing_words_path: typing.Optional[typing.Union[str, Path]] = None,
    number_repeated_words: bool = True,
):
    """Create pronunciation dictionary. Guess missing words if g2p model is available."""
    # Look up words
    missing_words: typing.Set[str] = set()

    # Look up each word
    with open(dictionary, "w") as dictionary_file:
        for word in vocabulary:
            word_phonemes = pronunciations.get(word)
            if not word_phonemes:
                # Add to missing word list
                _LOGGER.warning("Missing word '%s'", word)
                missing_words.add(word)
                continue

            # Write CMU format
            for i, phonemes in enumerate(word_phonemes):
                phoneme_str = " ".join(phonemes).strip()
                if (not number_repeated_words) or (i == 0):
                    # word
                    print(word, phoneme_str, file=dictionary_file)
                else:
                    # word(n)
                    print(f"{word}({i+1})", phoneme_str, file=dictionary_file)

        # Open missing words file
        missing_file: typing.Optional[typing.TextIO] = None
        if missing_words_path:
            missing_file = open(missing_words_path, "w")

        try:
            if missing_words:
                # Fail if no g2p model is available
                if not g2p_model:
                    raise MissingWordPronunciationsException(list(missing_words))

                if not phonetisaurus_apply:
                    # Find in PATH
                    phonetisaurus_apply = shutil.which("phonetisaurus-apply")
                    assert phonetisaurus_apply, "phonetisaurus-apply not found in PATH"

                # Guess word pronunciations
                _LOGGER.debug("Guessing pronunciations for %s", missing_words)
                guesses = guess_pronunciations(
                    missing_words,
                    g2p_model,
                    phonetisaurus_apply,
                    g2p_word_transform=g2p_word_transform,
                    num_guesses=1,
                )

                # Output is a pronunciation dictionary.
                # Append to existing dictionary file.
                for guess_word, guess_phonemes in guesses:
                    guess_phoneme_str = " ".join(guess_phonemes).strip()
                    print(guess_word, guess_phoneme_str, file=dictionary_file)

                    if missing_file:
                        print(guess_word, guess_phoneme_str, file=missing_file)

        finally:
            if missing_file:
                missing_file.close()
                _LOGGER.debug("Wrote missing words to %s", str(missing_words_path))


def guess_pronunciations(
    words: typing.Iterable[str],
    g2p_model: typing.Union[str, Path],
    phonetisaurus_apply: typing.Optional[typing.Union[str, Path]] = None,
    g2p_word_transform: typing.Optional[typing.Callable[[str], str]] = None,
    num_guesses: int = 1,
) -> typing.Iterable[typing.Tuple[str, typing.List[str]]]:
    """Guess phonetic pronunciations for words. Yields (word, phonemes) pairs."""
    if not phonetisaurus_apply:
        # Find in PATH
        phonetisaurus_apply = shutil.which("phonetisaurus-apply")
        assert phonetisaurus_apply, "phonetisaurus-apply not found in PATH"

    g2p_word_transform = g2p_word_transform or (lambda s: s)

    with tempfile.NamedTemporaryFile(mode="w") as wordlist_file:
        for word in words:
            word = g2p_word_transform(word)
            print(word, file=wordlist_file)

        wordlist_file.seek(0)
        g2p_command = [
            str(phonetisaurus_apply),
            "--model",
            str(g2p_model),
            "--word_list",
            wordlist_file.name,
            "--nbest",
            str(num_guesses),
        ]

        _LOGGER.debug(g2p_command)
        g2p_lines = subprocess.check_output(
            g2p_command, universal_newlines=True
        ).splitlines()

        # Output is a pronunciation dictionary.
        # Append to existing dictionary file.
        for line in g2p_lines:
            line = line.strip()
            if line:
                word, *phonemes = line.split()
                yield (word.strip(), phonemes)



"""Utilities to convert JSGF sentences to directed graphs."""
import base64
import gzip
import io
import math
import typing
from dataclasses import dataclass
from pathlib import Path

import networkx as nx

from .const import IntentsType, ReplacementsType, SentencesType
from .ini_jsgf import get_intent_counts, split_rules
from .jsgf import (
    Expression,
    RuleReference,
    Sentence,
    Sequence,
    SequenceType,
    SlotReference,
    Substitutable,
    Taggable,
    Word,
)
from .slots import split_slot_args

# -----------------------------------------------------------------------------


def expression_to_graph(
    expression: Expression,
    graph: nx.DiGraph,
    source_state: int,
    replacements: typing.Optional[ReplacementsType] = None,
    empty_substitution: bool = False,
    grammar_name: typing.Optional[str] = None,
    count_dict: typing.Optional[typing.Dict[Expression, int]] = None,
    rule_grammar: str = "",
    expand_slots: bool = True,
) -> int:
    """Insert JSGF expression into a graph. Return final state."""
    replacements = replacements or {}

    # Handle sequence substitution
    if isinstance(expression, Substitutable) and expression.substitution:
        # Ensure everything downstream outputs nothing
        empty_substitution = True

    # Handle tag begin
    if isinstance(expression, Taggable) and expression.tag:
        # Begin tag
        next_state = len(graph)
        tag = expression.tag.tag_text
        olabel = f"__begin__{tag}"
        label = f":{olabel}"
        graph.add_edge(
            source_state, next_state, ilabel="", olabel=maybe_pack(olabel), label=label
        )
        source_state = next_state

        if expression.tag.substitution:
            # Ensure everything downstream outputs nothing
            empty_substitution = True

    # Handle converters begin
    begin_converters: typing.List[str] = []
    if isinstance(expression, Taggable) and expression.tag:
        begin_converters.extend(reversed(expression.tag.converters))

    if isinstance(expression, Substitutable) and expression.converters:
        begin_converters.extend(reversed(expression.converters))

    # Create begin transitions for each converter (in reverse order)
    for converter_name in begin_converters:
        next_state = len(graph)
        olabel = f"__convert__{converter_name}"
        label = f"!{olabel}"
        graph.add_edge(
            source_state, next_state, ilabel="", olabel=maybe_pack(olabel), label=label
        )
        source_state = next_state

    if isinstance(expression, Sequence):
        # Group, optional, or alternative
        seq: Sequence = expression
        if seq.type == SequenceType.ALTERNATIVE:
            # Optional or alternative
            final_states = []
            for item in seq.items:
                # Branch alternatives from source state
                next_state = expression_to_graph(
                    item,
                    graph,
                    source_state,
                    replacements=replacements,
                    empty_substitution=empty_substitution,
                    grammar_name=grammar_name,
                    count_dict=count_dict,
                    rule_grammar=rule_grammar,
                    expand_slots=expand_slots,
                )
                final_states.append(next_state)

            # Connect all paths to final state
            next_state = len(graph)
            for final_state in final_states:
                graph.add_edge(final_state, next_state, ilabel="", olabel="", label="")

            source_state = next_state
        else:
            # Group
            next_state = source_state
            for item in seq.items:
                # Create sequence of states
                next_state = expression_to_graph(
                    item,
                    graph,
                    next_state,
                    replacements=replacements,
                    empty_substitution=empty_substitution,
                    grammar_name=grammar_name,
                    count_dict=count_dict,
                    rule_grammar=rule_grammar,
                    expand_slots=expand_slots,
                )

            source_state = next_state
    elif isinstance(expression, Word):
        # State for single word
        word: Word = expression
        next_state = len(graph)
        graph.add_node(next_state, word=word.text)

        if (word.substitution is None) and (not empty_substitution):
            # Single word input/output
            graph.add_edge(
                source_state,
                next_state,
                ilabel=word.text,
                olabel=word.text,
                label=word.text,
            )
            source_state = next_state
        else:
            # Loading edge
            graph.add_edge(
                source_state,
                next_state,
                ilabel=word.text,
                olabel="",
                label=f"{word.text}:",
            )

            source_state = next_state

            # Add word output(s)
            olabels = [word.text] if (word.substitution is None) else word.substitution
            if not empty_substitution:
                source_state = add_substitution(graph, olabels, source_state)
    elif isinstance(expression, RuleReference):
        # Reference to a local or remote rule
        rule_ref: RuleReference = expression
        if rule_ref.grammar_name:
            # Fully resolved rule name
            rule_name = f"{rule_ref.grammar_name}.{rule_ref.rule_name}"
            rule_grammar = rule_ref.grammar_name
        elif rule_grammar:
            # Nested rule
            rule_name = f"{rule_grammar}.{rule_ref.rule_name}"
        elif grammar_name:
            # Local rule
            rule_name = f"{grammar_name}.{rule_ref.rule_name}"
            rule_grammar = grammar_name
        else:
            # Unresolved rule name
            rule_name = rule_ref.rule_name

        # Surround with <>
        rule_name_brackets = f"<{rule_name}>"
        rule_replacements = replacements.get(rule_name_brackets)
        assert rule_replacements, f"Missing rule {rule_name}"

        rule_body = next(iter(rule_replacements))
        assert isinstance(rule_body, Sentence), f"Invalid rule {rule_name}: {rule_body}"
        source_state = expression_to_graph(
            rule_body,
            graph,
            source_state,
            replacements=replacements,
            empty_substitution=empty_substitution,
            grammar_name=grammar_name,
            count_dict=count_dict,
            rule_grammar=rule_grammar,
            expand_slots=expand_slots,
        )

    elif isinstance(expression, SlotReference):
        # Reference to slot values
        slot_ref: SlotReference = expression

        # Prefix with $
        slot_name = "$" + slot_ref.slot_name

        if expand_slots:
            slot_values = replacements.get(slot_name)
            assert slot_values, f"Missing slot {slot_name}"

            # Interpret as alternative
            slot_seq = Sequence(type=SequenceType.ALTERNATIVE, items=list(slot_values))
            source_state = expression_to_graph(
                slot_seq,
                graph,
                source_state,
                replacements=replacements,
                empty_substitution=(empty_substitution or bool(slot_ref.substitution)),
                grammar_name=grammar_name,
                count_dict=count_dict,
                rule_grammar=rule_grammar,
                expand_slots=expand_slots,
            )

        # Emit __source__ with slot name (no arguments)
        slot_name_noargs = split_slot_args(slot_ref.slot_name)[0]
        next_state = len(graph)
        olabel = f"__source__{slot_name_noargs}"
        graph.add_edge(
            source_state, next_state, ilabel="", olabel=olabel, label=maybe_pack(olabel)
        )
        source_state = next_state

    # Handle sequence substitution
    if isinstance(expression, Substitutable) and expression.substitution:
        # Output substituted word(s)
        source_state = add_substitution(graph, expression.substitution, source_state)

    # Handle converters end
    end_converters: typing.List[str] = []
    if isinstance(expression, Substitutable) and expression.converters:
        end_converters.extend(expression.converters)

    if isinstance(expression, Taggable) and expression.tag:
        end_converters.extend(expression.tag.converters)

    # Create end transitions for each converter
    for converter_name in end_converters:
        next_state = len(graph)
        olabel = f"__converted__{converter_name}"
        label = f"!{olabel}"
        graph.add_edge(
            source_state, next_state, ilabel="", olabel=maybe_pack(olabel), label=label
        )
        source_state = next_state

    # Handle tag end
    if isinstance(expression, Taggable) and expression.tag:
        # Handle tag substitution
        if expression.tag.substitution:
            # Output substituted word(s)
            source_state = add_substitution(
                graph, expression.tag.substitution, source_state
            )

        # End tag
        next_state = len(graph)
        tag = expression.tag.tag_text
        olabel = f"__end__{tag}"
        label = f":{olabel}"
        graph.add_edge(
            source_state, next_state, ilabel="", olabel=maybe_pack(olabel), label=label
        )
        source_state = next_state

    return source_state


def add_substitution(
    graph: nx.DiGraph,
    substitution: typing.Union[str, typing.List[str]],
    source_state: int,
) -> int:
    """Add substitution token sequence to graph."""
    if isinstance(substitution, str):
        substitution = [substitution]

    for olabel in substitution:
        next_state = len(graph)
        graph.add_edge(
            source_state,
            next_state,
            ilabel="",
            olabel=maybe_pack(olabel),
            label=f":{olabel}",
        )

        source_state = next_state

    return source_state


def maybe_pack(olabel: str) -> str:
    """Pack output label as base64 if it contains whitespace."""
    if " " in olabel:
        return "__unpack__" + base64.encodebytes(olabel.encode()).decode().strip()

    return olabel


# -----------------------------------------------------------------------------


def intents_to_graph(
    intents: IntentsType,
    replacements: typing.Optional[ReplacementsType] = None,
    add_intent_weights: bool = True,
    exclude_slots_from_counts: bool = True,
) -> nx.DiGraph:
    """Convert sentences/rules grouped by intent into a directed graph."""
    sentences, replacements = split_rules(intents, replacements)
    return sentences_to_graph(
        sentences,
        replacements=replacements,
        add_intent_weights=add_intent_weights,
        exclude_slots_from_counts=exclude_slots_from_counts,
    )


def sentences_to_graph(
    sentences: SentencesType,
    replacements: typing.Optional[ReplacementsType] = None,
    add_intent_weights: bool = True,
    exclude_slots_from_counts: bool = True,
    expand_slots: bool = True,
) -> nx.DiGraph:
    """Convert sentences grouped by intent into a directed graph."""
    num_intents = len(sentences)
    intent_weights: typing.Dict[str, float] = {}
    count_dict: typing.Optional[typing.Dict[Expression, int]] = None

    if add_intent_weights:
        # Count number of posssible sentences per intent
        intent_counts = get_intent_counts(
            sentences,
            replacements,
            exclude_slots=exclude_slots_from_counts,
            count_dict=count_dict,
        )

        # Fix zero counts
        for intent_name in intent_counts:
            intent_counts[intent_name] = max(intent_counts[intent_name], 1)

        num_sentences_lcm = lcm(*intent_counts.values())
        intent_weights = {
            intent_name: (
                num_sentences_lcm // max(intent_counts.get(intent_name, 1), 1)
            )
            for intent_name in sentences
        }

        # Normalize
        weight_sum = max(sum(intent_weights.values()), 1)
        for intent_name in intent_weights:
            intent_weights[intent_name] /= weight_sum
    else:
        intent_counts = {}

    # Create initial graph
    graph: nx.DiGraph = nx.DiGraph()
    root_state: int = 0
    graph.add_node(root_state, start=True)
    final_states: typing.List[int] = []

    for intent_name, intent_sentences in sentences.items():
        # Branch off for each intent from start state
        intent_state = len(graph)
        olabel = f"__label__{intent_name}"
        label = f":{olabel}"

        edge_kwargs = {}
        if add_intent_weights and (num_intents > 1):
            edge_kwargs["sentence_count"] = intent_counts.get(intent_name, 1)
            edge_kwargs["weight"] = intent_weights.get(intent_name, 0)

        graph.add_edge(
            root_state,
            intent_state,
            ilabel="",
            olabel=olabel,
            label=label,
            **edge_kwargs,
        )

        for sentence in intent_sentences:
            # Insert all sentences for this intent
            next_state = expression_to_graph(  # type: ignore
                sentence,
                graph,
                intent_state,
                replacements=replacements,
                grammar_name=intent_name,
                count_dict=count_dict,
                expand_slots=expand_slots,
            )
            final_states.append(next_state)

    # Create final state and join all sentences to it
    final_state = len(graph)
    graph.add_node(final_state, final=True)

    for next_state in final_states:
        graph.add_edge(next_state, final_state, ilabel="", olabel="", label="")

    return graph


# -----------------------------------------------------------------------------


def graph_to_json(graph: nx.DiGraph) -> typing.Dict[str, typing.Any]:
    """Convert to dict suitable for JSON serialization."""
    return nx.readwrite.json_graph.node_link_data(graph)


def json_to_graph(json_dict: typing.Dict[str, typing.Any]) -> nx.DiGraph:
    """Convert from deserialized JSON dict to graph."""
    return nx.readwrite.json_graph.node_link_graph(json_dict)


def graph_to_gzip_pickle(graph: nx.DiGraph, out_file: typing.BinaryIO):
    """Convert to binary gzip pickle format."""
    with gzip.GzipFile(fileobj=out_file, mode="wb") as graph_gzip:
        nx.readwrite.gpickle.write_gpickle(graph, graph_gzip)


def gzip_pickle_to_graph(in_file: typing.BinaryIO) -> nx.DiGraph:
    """Convert from binary gzip pickle format."""
    with gzip.GzipFile(fileobj=in_file, mode="rb") as graph_gzip:
        return nx.readwrite.gpickle.read_gpickle(graph_gzip)


# -----------------------------------------------------------------------------


@dataclass
class GraphFsts:
    """Result from graph_to_fsts."""

    intent_fsts: typing.Dict[str, str]
    symbols: typing.Dict[str, int]
    input_symbols: typing.Dict[str, int]
    output_symbols: typing.Dict[str, int]


def graph_to_fsts(
    graph: nx.DiGraph,
    eps="<eps>",
    weight_key="weight",
    default_weight=0,
    intent_filter: typing.Optional[typing.Callable[[str], bool]] = None,
) -> GraphFsts:
    """Convert graph to OpenFST text format, one per intent."""
    intent_fsts: typing.Dict[str, str] = {}
    symbols: typing.Dict[str, int] = {eps: 0}
    input_symbols: typing.Dict[str, int] = {}
    output_symbols: typing.Dict[str, int] = {}
    n_data = graph.nodes(data=True)

    # start state
    start_node: int = next(n for n, data in n_data if data.get("start"))

    for _, intent_node, edge_data in graph.edges(start_node, data=True):
        intent_name: str = edge_data["olabel"][9:]

        # Filter intents by name
        if intent_filter and not intent_filter(intent_name):
            continue

        final_states: typing.Set[int] = set()
        state_map: typing.Dict[int, int] = {}

        with io.StringIO() as intent_file:
            # Transitions
            for edge in nx.edge_bfs(graph, intent_node):
                edge_data = graph.edges[edge]
                from_node, to_node = edge

                # Map states starting from 0
                from_state = state_map.get(from_node, len(state_map))
                state_map[from_node] = from_state

                to_state = state_map.get(to_node, len(state_map))
                state_map[to_node] = to_state

                # Get input/output labels.
                # Empty string indicates epsilon transition (eps)
                ilabel = edge_data.get("ilabel", "") or eps
                olabel = edge_data.get("olabel", "") or eps

                # Map labels (symbols) to integers
                isymbol = symbols.get(ilabel, len(symbols))
                symbols[ilabel] = isymbol
                input_symbols[ilabel] = isymbol

                osymbol = symbols.get(olabel, len(symbols))
                symbols[olabel] = osymbol
                output_symbols[olabel] = osymbol

                if weight_key:
                    weight = edge_data.get(weight_key, default_weight)
                    print(
                        f"{from_state} {to_state} {ilabel} {olabel} {weight}",
                        file=intent_file,
                    )
                else:
                    # No weight
                    print(
                        f"{from_state} {to_state} {ilabel} {olabel}", file=intent_file
                    )

                # Check if final state
                if n_data[from_node].get("final", False):
                    final_states.add(from_state)

                if n_data[to_node].get("final", False):
                    final_states.add(to_state)

            # Record final states
            for final_state in final_states:
                print(final_state, file=intent_file)

            intent_fsts[intent_name] = intent_file.getvalue()

    return GraphFsts(
        intent_fsts=intent_fsts,
        symbols=symbols,
        input_symbols=input_symbols,
        output_symbols=output_symbols,
    )


# -----------------------------------------------------------------------------


@dataclass
class GraphFst:
    """Result from graph_to_fst."""

    intent_fst: str
    symbols: typing.Dict[str, int]
    input_symbols: typing.Dict[str, int]
    output_symbols: typing.Dict[str, int]

    def write_fst(
        self,
        fst_text_path: typing.Union[str, Path],
        isymbols_path: typing.Union[str, Path],
        osymbols_path: typing.Union[str, Path],
    ):
        """Write FST text and symbol files."""
        # Write FST
        Path(fst_text_path).write_text(self.intent_fst)

        # Write input symbols
        with open(isymbols_path, "w") as isymbols_file:
            # pylint: disable=E1101
            for symbol, num in self.input_symbols.items():
                print(symbol, num, file=isymbols_file)

        # Write output symbols
        with open(osymbols_path, "w") as osymbols_file:
            # pylint: disable=E1101
            for symbol, num in self.output_symbols.items():
                print(symbol, num, file=osymbols_file)


def graph_to_fst(
    graph: nx.DiGraph,
    eps="<eps>",
    weight_key="weight",
    default_weight=0,
    intent_filter: typing.Optional[typing.Callable[[str], bool]] = None,
) -> GraphFst:
    """Convert graph to OpenFST text format."""
    symbols: typing.Dict[str, int] = {eps: 0}
    input_symbols: typing.Dict[str, int] = {}
    output_symbols: typing.Dict[str, int] = {}
    n_data = graph.nodes(data=True)

    # start state
    start_node: int = next(n for n, data in n_data if data.get("start"))

    # Generate FST text
    with io.StringIO() as fst_file:
        final_states: typing.Set[int] = set()
        state_map: typing.Dict[int, int] = {}

        # Transitions
        for _, intent_node, intent_edge_data in graph.edges(start_node, data=True):
            intent_olabel: str = intent_edge_data["olabel"]
            intent_name: str = intent_olabel[9:]

            # Filter intents by name
            if intent_filter and not intent_filter(intent_name):
                continue

            assert (
                " " not in intent_olabel
            ), f"Output symbol cannot contain whitespace: {intent_olabel}"

            # Map states starting from 0
            from_state = state_map.get(start_node, len(state_map))
            state_map[start_node] = from_state

            to_state = state_map.get(intent_node, len(state_map))
            state_map[intent_node] = to_state

            # Map labels (symbols) to integers
            isymbol = symbols.get(eps, len(symbols))
            symbols[eps] = isymbol
            input_symbols[eps] = isymbol

            osymbol = symbols.get(intent_olabel, len(symbols))
            symbols[intent_olabel] = osymbol
            output_symbols[intent_olabel] = osymbol

            if weight_key:
                weight = intent_edge_data.get(weight_key, default_weight)
                print(
                    f"{from_state} {to_state} {eps} {intent_olabel} {weight}",
                    file=fst_file,
                )
            else:
                # No weight
                print(f"{from_state} {to_state} {eps} {intent_olabel}", file=fst_file)

            # Add intent sub-graphs
            for edge in nx.edge_bfs(graph, intent_node):
                edge_data = graph.edges[edge]
                from_node, to_node = edge

                # Get input/output labels.
                # Empty string indicates epsilon transition (eps)
                ilabel = edge_data.get("ilabel", "") or eps
                olabel = edge_data.get("olabel", "") or eps

                # Check for whitespace
                assert (
                    " " not in ilabel
                ), f"Input symbol cannot contain whitespace: {ilabel}"

                assert (
                    " " not in olabel
                ), f"Output symbol cannot contain whitespace: {olabel}"

                # Map states starting from 0
                from_state = state_map.get(from_node, len(state_map))
                state_map[from_node] = from_state

                to_state = state_map.get(to_node, len(state_map))
                state_map[to_node] = to_state

                # Map labels (symbols) to integers
                isymbol = symbols.get(ilabel, len(symbols))
                symbols[ilabel] = isymbol
                input_symbols[ilabel] = isymbol

                osymbol = symbols.get(olabel, len(symbols))
                symbols[olabel] = osymbol
                output_symbols[olabel] = osymbol

                if weight_key:
                    weight = edge_data.get(weight_key, default_weight)
                    print(
                        f"{from_state} {to_state} {ilabel} {olabel} {weight}",
                        file=fst_file,
                    )
                else:
                    # No weight
                    print(f"{from_state} {to_state} {ilabel} {olabel}", file=fst_file)

                # Check if final state
                if n_data[from_node].get("final", False):
                    final_states.add(from_state)

                if n_data[to_node].get("final", False):
                    final_states.add(to_state)

        # Record final states
        for final_state in final_states:
            print(final_state, file=fst_file)

        return GraphFst(
            intent_fst=fst_file.getvalue(),
            symbols=symbols,
            input_symbols=input_symbols,
            output_symbols=output_symbols,
        )


# -----------------------------------------------------------------------------


def lcm(*nums: int) -> int:
    """Returns the least common multiple of the given integers"""
    if nums:
        nums_lcm = nums[0]
        for n in nums[1:]:
            nums_lcm = (nums_lcm * n) // math.gcd(nums_lcm, n)

        return nums_lcm

    return 1


# -----------------------------------------------------------------------------


def get_start_end_nodes(
    graph: nx.DiGraph,
) -> typing.Tuple[typing.Optional[int], typing.Optional[int]]:
    """Return start/end nodes in graph"""
    n_data = graph.nodes(data=True)
    start_node = None
    end_node = None

    for node, data in n_data:
        if data.get("start", False):
            start_node = node
        elif data.get("final", False):
            end_node = node

        if (start_node is not None) and (end_node is not None):
            break

    return (start_node, end_node)



"""Utility methods for rhasspynlu"""
import dataclasses
import itertools
import typing


def pairwise(iterable: typing.Iterable[typing.Any]):
    """s -> (s0,s1), (s1,s2), (s2,s3), ..."""
    a, b = itertools.tee(iterable)
    return zip(a, itertools.islice(b, 1, None))


def only_fields(
    cls, message_dict: typing.Dict[str, typing.Any]
) -> typing.Dict[str, typing.Any]:
    """Return dict with only valid fields."""
    if dataclasses.is_dataclass(cls):
        field_names = set(f.name for f in dataclasses.fields(cls))
        valid_fields = set(message_dict.keys()).intersection(field_names)
        return {key: message_dict[key] for key in valid_fields}

    return message_dict



"""Command-line utility for rhasspynlu"""
import argparse
import io
import json
import logging
import os
import sys

from . import graph_to_fst, graph_to_json, intents_to_graph, parse_ini
from .fsticuffs import sample_by_intent

_LOGGER = logging.getLogger(__name__)


def main():
    """Main method"""
    parser = argparse.ArgumentParser("rhasspynlu")
    parser.add_argument("sentences_file", nargs="*", help="Sentences ini files")
    parser.add_argument(
        "--fst", action="store_true", help="Output FST text instead of JSON"
    )
    parser.add_argument("--fst-text", default="fst.txt", help="Path to FST text file")
    parser.add_argument(
        "--fst-isymbols",
        default="fst.isymbols.txt",
        help="Path to FST input symbols file",
    )
    parser.add_argument(
        "--fst-osymbols",
        default="fst.osymbols.txt",
        help="Path to FST output symbols file",
    )
    parser.add_argument(
        "--sample", type=int, help="Generate sample sentences (0 = all)"
    )
    parser.add_argument(
        "--debug", action="store_true", help="Print DEBUG messages to console"
    )
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.INFO)

    _LOGGER.debug(args)

    # Read sentences ini from files or stdin
    try:
        if args.sentences_file:
            with io.StringIO() as ini_file:
                for sentences_path in args.sentences_file:
                    _LOGGER.debug("Reading %s", sentences_path)

                    with open(sentences_path, "r") as sentences_file:
                        ini_file.write(sentences_file.read())
                        print("", file=ini_file)

                ini_text = ini_file.getvalue()
        else:
            if os.isatty(sys.stdin.fileno()):
                print("Reading sentences ini from stdin...", file=sys.stderr)

            ini_text = sys.stdin.read()

        _LOGGER.debug("Parsing sentences")
        intents = parse_ini(ini_text)

        _LOGGER.debug("Converting to graph")
        graph = intents_to_graph(intents)

        if args.fst:
            _LOGGER.debug("Converting to FST")
            graph_fst = graph_to_fst(graph)

            _LOGGER.debug(
                "Writing %s, %s, %s",
                args.fst_text,
                args.fst_isymbols,
                args.fst_osymbols,
            )

            graph_fst.write_fst(args.fst_text, args.fst_isymbols, args.fst_osymbols)
        elif args.sample is not None:
            if args.sample > 0:
                num_samples = args.sample
                _LOGGER.debug("Generating %s sample(s)", args.sample)
            else:
                num_samples = None
                _LOGGER.debug("Generating all sentences")

            sentences_by_intent = sample_by_intent(graph, num_samples)

            # Output JSON
            json.dump(
                {
                    intent: [r.asdict() for r in recognitions]
                    for intent, recognitions in sentences_by_intent.items()
                },
                sys.stdout,
            )
        else:
            # Output JSON
            _LOGGER.debug("Writing to stdout")
            graph_dict = graph_to_json(graph)

            json.dump(graph_dict, sys.stdout)
    except KeyboardInterrupt:
        pass


# -----------------------------------------------------------------------------

if __name__ == "__main__":
    main()



"""Setup script for rhasspynlu"""
import os

import setuptools

this_dir = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_dir, "README.md"), "r") as readme_file:
    long_description = readme_file.read()

with open("requirements.txt", "r") as requirements_file:
    requirements = requirements_file.read().splitlines()

with open("VERSION", "r") as version_file:
    version = version_file.read().strip()

setuptools.setup(
    name="rhasspy-nlu",
    version=version,
    author="Michael Hansen",
    author_email="hansen.mike@gmail.com",
    url="https://github.com/rhasspy/rhasspy-nlu",
    packages=setuptools.find_packages(),
    package_data={"rhasspynlu": ["py.typed"]},
    install_requires=requirements,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
)



"""Test cases for recognition functions."""
import unittest
from pathlib import Path

from rhasspynlu.fsticuffs import recognize
from rhasspynlu.ini_jsgf import parse_ini
from rhasspynlu.intent import Entity, Intent, Recognition
from rhasspynlu.jsgf import Sentence
from rhasspynlu.jsgf_graph import intents_to_graph


class StrictTestCase(unittest.TestCase):
    """Strict recognition test cases."""

    def test_single_sentence(self):
        """Single intent, single sentence."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test
        """
        )

        graph = intents_to_graph(intents)

        # Exact
        recognitions = zero_times(recognize("this is a test", graph, fuzzy=False))
        print(recognitions)
        self.assertEqual(
            recognitions,
            [
                Recognition(
                    intent=Intent(name="TestIntent", confidence=1.0),
                    text="this is a test",
                    raw_text="this is a test",
                    tokens=["this", "is", "a", "test"],
                    raw_tokens=["this", "is", "a", "test"],
                )
            ],
        )

        # Too many tokens (lower confidence)
        recognitions = zero_times(recognize("this is a bad test", graph, fuzzy=False))
        self.assertFalse(recognitions)

        # Too few tokens (failure)
        recognitions = zero_times(recognize("this is a", graph, fuzzy=False))
        self.assertFalse(recognitions)

    def test_multiple_sentences(self):
        """Identical sentences from two different intents."""
        intents = parse_ini(
            """
        [TestIntent1]
        this is a test

        [TestIntent2]
        this is a test
        """
        )

        graph = intents_to_graph(intents)

        # Should produce a recognition for each intent
        recognitions = zero_times(recognize("this is a test", graph, fuzzy=False))
        self.assertEqual(len(recognitions), 2)
        self.assertIn(
            Recognition(
                intent=Intent(name="TestIntent1", confidence=1.0),
                text="this is a test",
                raw_text="this is a test",
                tokens=["this", "is", "a", "test"],
                raw_tokens=["this", "is", "a", "test"],
            ),
            recognitions,
        )
        self.assertIn(
            Recognition(
                intent=Intent(name="TestIntent2", confidence=1.0),
                text="this is a test",
                raw_text="this is a test",
                tokens=["this", "is", "a", "test"],
                raw_tokens=["this", "is", "a", "test"],
            ),
            recognitions,
        )

    def test_stop_words(self):
        """Check sentence with stop words."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test
        """
        )

        graph = intents_to_graph(intents)

        # Failure without stop words
        recognitions = zero_times(recognize("this is a abcd test", graph, fuzzy=False))
        self.assertFalse(recognitions)

        # Success with stop words
        recognitions = zero_times(
            recognize("this is a abcd test", graph, stop_words={"abcd"}, fuzzy=False)
        )
        self.assertEqual(
            recognitions,
            [
                Recognition(
                    intent=Intent(name="TestIntent", confidence=1.0),
                    text="this is a test",
                    raw_text="this is a test",
                    tokens=["this", "is", "a", "test"],
                    raw_tokens=["this", "is", "a", "test"],
                )
            ],
        )

    def test_converters(self):
        """Check sentence with converters."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test!upper ten:10!int!square
        """
        )

        graph = intents_to_graph(intents)

        # Should upper-case "test" and convert "ten" -> 10 -> 100
        recognitions = zero_times(
            recognize(
                "this is a test ten",
                graph,
                fuzzy=False,
                extra_converters={"square": lambda *args: [x ** 2 for x in args]},
            )
        )
        self.assertEqual(
            recognitions,
            [
                Recognition(
                    intent=Intent(name="TestIntent", confidence=1.0),
                    text="this is a TEST 100",
                    raw_text="this is a test ten",
                    tokens=["this", "is", "a", "TEST", 100],
                    raw_tokens=["this", "is", "a", "test", "ten"],
                )
            ],
        )

    def test_converter_args(self):
        """Check converter with arguments."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test ten:10!int!pow,3
        """
        )

        graph = intents_to_graph(intents)

        def pow_converter(*args, converter_args=None):
            exponent = int(converter_args[0]) if converter_args else 1
            return [x ** exponent for x in args]

        # Should convert "ten" -> 10 -> 1000
        recognitions = zero_times(
            recognize(
                "this is a test ten",
                graph,
                fuzzy=False,
                extra_converters={"pow": pow_converter},
            )
        )
        self.assertEqual(
            recognitions,
            [
                Recognition(
                    intent=Intent(name="TestIntent", confidence=1.0),
                    text="this is a test 1000",
                    raw_text="this is a test ten",
                    tokens=["this", "is", "a", "test", 1000],
                    raw_tokens=["this", "is", "a", "test", "ten"],
                )
            ],
        )


# -----------------------------------------------------------------------------


class FuzzyTestCase(unittest.TestCase):
    """Fuzzy recognition test cases."""

    def test_single_sentence(self):
        """Single intent, single sentence."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test
        """
        )

        graph = intents_to_graph(intents)

        # Exact
        recognitions = zero_times(recognize("this is a test", graph))
        self.assertEqual(
            recognitions,
            [
                Recognition(
                    intent=Intent(name="TestIntent", confidence=1.0),
                    text="this is a test",
                    raw_text="this is a test",
                    tokens=["this", "is", "a", "test"],
                    raw_tokens=["this", "is", "a", "test"],
                )
            ],
        )

        # Too many tokens (lower confidence)
        recognitions = zero_times(recognize("this is a bad test", graph))
        self.assertEqual(
            recognitions,
            [
                Recognition(
                    intent=Intent(name="TestIntent", confidence=float(1 - 1 / 4)),
                    text="this is a test",
                    raw_text="this is a test",
                    tokens=["this", "is", "a", "test"],
                    raw_tokens=["this", "is", "a", "test"],
                )
            ],
        )

        # Too few tokens (failure)
        recognitions = zero_times(recognize("this is a", graph))
        self.assertFalse(recognitions)

    def test_multiple_sentences(self):
        """Identical sentences from two different intents."""
        intents = parse_ini(
            """
        [TestIntent1]
        this is a test

        [TestIntent2]
        this is a test
        """
        )

        graph = intents_to_graph(intents)

        # Should produce a recognition for each intent
        recognitions = zero_times(recognize("this is a test", graph))
        self.assertEqual(len(recognitions), 2)
        self.assertIn(
            Recognition(
                intent=Intent(name="TestIntent1", confidence=1.0),
                text="this is a test",
                raw_text="this is a test",
                tokens=["this", "is", "a", "test"],
                raw_tokens=["this", "is", "a", "test"],
            ),
            recognitions,
        )
        self.assertIn(
            Recognition(
                intent=Intent(name="TestIntent2", confidence=1.0),
                text="this is a test",
                raw_text="this is a test",
                tokens=["this", "is", "a", "test"],
                raw_tokens=["this", "is", "a", "test"],
            ),
            recognitions,
        )

    def test_intent_filter(self):
        """Identical sentences from two different intents with filter."""
        intents = parse_ini(
            """
        [TestIntent1]
        this is a test

        [TestIntent2]
        this is a test
        """
        )

        graph = intents_to_graph(intents)

        def intent_filter(name):
            return name == "TestIntent1"

        # Should produce a recognition for first intent only
        recognitions = zero_times(
            recognize("this is a test", graph, intent_filter=intent_filter)
        )
        self.assertEqual(
            recognitions,
            [
                Recognition(
                    intent=Intent(name="TestIntent1", confidence=1.0),
                    text="this is a test",
                    raw_text="this is a test",
                    tokens=["this", "is", "a", "test"],
                    raw_tokens=["this", "is", "a", "test"],
                )
            ],
        )

    def test_stop_words(self):
        """Check sentence with stop words."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test
        """
        )

        graph = intents_to_graph(intents)

        # Lower confidence with no stop words
        recognitions = zero_times(recognize("this is a abcd test", graph))
        self.assertEqual(len(recognitions), 1)
        self.assertEqual(recognitions[0].intent.confidence, 1 - (1 / 4))

        # Higher confidence with stop words
        recognitions = zero_times(
            recognize("this is a abcd test", graph, stop_words={"abcd"})
        )
        self.assertEqual(
            recognitions,
            [
                Recognition(
                    intent=Intent(name="TestIntent", confidence=float(1 - (0.1 / 4))),
                    text="this is a test",
                    raw_text="this is a test",
                    tokens=["this", "is", "a", "test"],
                    raw_tokens=["this", "is", "a", "test"],
                )
            ],
        )

    def test_rules(self):
        """Make sure local and remote rules work."""
        intents = parse_ini(
            """
        [Intent1]
        rule = a test
        this is <rule>

        [Intent2]
        rule = this is
        <rule> <Intent1.rule>
        """
        )

        graph = intents_to_graph(intents)

        # Lower confidence with no stop words
        recognitions = zero_times(recognize("this is a test", graph))
        self.assertEqual(
            recognitions,
            [
                Recognition(
                    intent=Intent(name="Intent1", confidence=1.0),
                    text="this is a test",
                    raw_text="this is a test",
                    tokens=["this", "is", "a", "test"],
                    raw_tokens=["this", "is", "a", "test"],
                ),
                Recognition(
                    intent=Intent(name="Intent2", confidence=1.0),
                    text="this is a test",
                    raw_text="this is a test",
                    tokens=["this", "is", "a", "test"],
                    raw_tokens=["this", "is", "a", "test"],
                ),
            ],
        )


# -----------------------------------------------------------------------------


class TimerTestCase(unittest.TestCase):
    """Test cases for timer example."""

    def setUp(self):
        # Load timer example
        self.graph = intents_to_graph(parse_ini(Path("etc/timer.ini")))

    def test_strict_simple(self):
        """Check exact parse."""
        recognitions = zero_times(
            recognize("set a timer for ten minutes", self.graph, fuzzy=False)
        )
        self.assertEqual(len(recognitions), 1)
        recognition = recognitions[0]
        self.assertTrue(recognition.intent)

        self.assertEqual(
            recognition.entities,
            [
                Entity(
                    entity="minutes",
                    value="10",
                    raw_value="ten",
                    start=16,
                    raw_start=16,
                    raw_end=19,
                    end=18,
                    tokens=["10"],
                    raw_tokens=["ten"],
                )
            ],
        )


# -----------------------------------------------------------------------------


class MiscellaneousTestCase(unittest.TestCase):
    """Miscellaneous test cases."""

    def test_optional_entity(self):
        """Ensure entity inside optional is recognized."""
        ini_text = """
        [playBook]
        read me ($audio-book-name){book} in [the] [($assistant-zones){zone}]
        """

        replacements = {
            "$audio-book-name": [Sentence.parse("the hound of the baskervilles")],
            "$assistant-zones": [Sentence.parse("bedroom")],
        }

        graph = intents_to_graph(parse_ini(ini_text), replacements)

        recognitions = zero_times(
            recognize(
                "read me the hound of the baskervilles in the bedroom",
                graph,
                fuzzy=False,
            )
        )
        self.assertEqual(len(recognitions), 1)
        recognition = recognitions[0]
        self.assertTrue(recognition.intent)

        entities = {e.entity: e for e in recognition.entities}
        self.assertIn("book", entities)
        book = entities["book"]
        self.assertEqual(book.value, "the hound of the baskervilles")

        self.assertIn("zone", entities)
        zone = entities["zone"]
        self.assertEqual(zone.value, "bedroom")

    def test_converters_in_entities(self):
        """Check sentence with converters inside an entity."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test (ten:10!int){number}
        """
        )

        graph = intents_to_graph(intents)

        # ten -> 10 (int)
        recognitions = zero_times(recognize("this is a test ten", graph, fuzzy=False))

        self.assertEqual(len(recognitions), 1)
        recognition = recognitions[0]
        self.assertTrue(recognition.intent)

        entities = {e.entity: e for e in recognition.entities}
        self.assertIn("number", entities)
        number = entities["number"]
        self.assertEqual(number.value, 10)

    def test_entity_converter(self):
        """Check sentence with an entity converter."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test (four: point: two:4.2){number!float}
        """
        )

        graph = intents_to_graph(intents)

        # "four point two" -> 4.2
        recognitions = zero_times(
            recognize("this is a test four point two", graph, fuzzy=False)
        )

        self.assertEqual(len(recognitions), 1)
        recognition = recognitions[0]
        self.assertTrue(recognition.intent)

        entities = {e.entity: e for e in recognition.entities}
        self.assertIn("number", entities)
        number = entities["number"]
        self.assertEqual(number.value, 4.2)

    def test_entity_converters_both(self):
        """Check sentence with an entity converter and a converter inside the entity."""
        intents = parse_ini(
            """
        [TestIntent]
        this is a test (four:4 point: two:2){number!floatify}
        """
        )

        graph = intents_to_graph(intents)

        # "four two" -> 4.2
        recognitions = zero_times(
            recognize(
                "this is a test four point two",
                graph,
                fuzzy=False,
                extra_converters={"floatify": lambda a, b: [float(f"{a}.{b}")]},
            )
        )

        self.assertEqual(len(recognitions), 1)
        recognition = recognitions[0]
        self.assertTrue(recognition.intent)

        entities = {e.entity: e for e in recognition.entities}
        self.assertIn("number", entities)
        number = entities["number"]
        self.assertEqual(number.value, 4.2)

    def test_sequence_converters(self):
        """Check sentence with sequence converters."""
        intents = parse_ini(
            """
        [TestIntent]
        this (is a test)!upper
        """
        )

        graph = intents_to_graph(intents)

        # Should upper-case "is a test"
        recognitions = zero_times(recognize("this is a test", graph, fuzzy=False))
        self.assertEqual(
            recognitions,
            [
                Recognition(
                    intent=Intent(name="TestIntent", confidence=1.0),
                    text="this IS A TEST",
                    raw_text="this is a test",
                    tokens=["this", "IS", "A", "TEST"],
                    raw_tokens=["this", "is", "a", "test"],
                )
            ],
        )

    def test_slot_sequence_replacement(self):
        """Ensure word sequences in slots can be replaced."""
        ini_text = """
        [PlayMusic]
        play me ($music_genre){genre}
        """

        replacements = {
            "$music_genre": [
                Sentence.parse("(rock | hard rock):(Hard Rock)"),
                Sentence.parse("classical:(Classical Music)"),
            ]
        }
        graph = intents_to_graph(parse_ini(ini_text), replacements)

        for text in ["play me rock", "play me hard rock"]:
            recognitions = zero_times(recognize(text, graph, fuzzy=False))
            self.assertEqual(len(recognitions), 1)
            recognition = recognitions[0]
            self.assertIsNotNone(recognition.intent)

            # Check sequence substitution
            self.assertEqual(recognition.text, "play me Hard Rock")

            # Check entity source
            self.assertEqual(len(recognition.entities), 1)
            genre = recognition.entities[0]
            self.assertEqual(genre.source, "music_genre")

        recognitions = zero_times(recognize("play me classical", graph, fuzzy=False))
        self.assertEqual(len(recognitions), 1)
        recognition = recognitions[0]
        self.assertIsNotNone(recognition.intent)

        # Check sequence substitution
        self.assertEqual(recognition.text, "play me Classical Music")

        # Check entity source
        self.assertEqual(len(recognition.entities), 1)
        genre = recognition.entities[0]
        self.assertEqual(genre.source, "music_genre")

    def test_word_case_preservation(self):
        """Ensure word casing is preserved in raw text."""
        ini_text = """
        [TestIntent]
        this is a (test){value}
        """

        graph = intents_to_graph(parse_ini(ini_text))

        recognitions = zero_times(
            recognize("this is a TEST", graph, fuzzy=False, word_transform=str.lower)
        )
        self.assertEqual(len(recognitions), 1)
        recognition = recognitions[0]
        self.assertIsNotNone(recognition.intent)

        # Check sequence substitution
        self.assertEqual(recognition.text, "this is a test")
        self.assertEqual(recognition.raw_text, "this is a TEST")

        self.assertEqual(len(recognition.entities), 1)
        value = recognition.entities[0]
        self.assertEqual(value.entity, "value")
        self.assertEqual(value.value, "test")
        self.assertEqual(value.raw_value, "TEST")

    def test_slot_case_preservation(self):
        """Ensure word casing is preserved in raw text."""
        ini_text = """
        [TestIntent]
        this is a ($test){value}
        """

        replacements = {"$test": [Sentence.parse("Bar")]}
        graph = intents_to_graph(parse_ini(ini_text), replacements)

        recognitions = zero_times(
            recognize("this is a bar", graph, fuzzy=False, word_transform=str.lower)
        )
        self.assertEqual(len(recognitions), 1)
        recognition = recognitions[0]
        self.assertIsNotNone(recognition.intent)

        # Check sequence substitution
        self.assertEqual(recognition.text, "this is a Bar")
        self.assertEqual(recognition.raw_text, "this is a bar")

        self.assertEqual(len(recognition.entities), 1)
        value = recognition.entities[0]
        self.assertEqual(value.entity, "value")
        self.assertEqual(value.value, "Bar")
        self.assertEqual(value.raw_value, "bar")
        self.assertEqual(value.source, "test")

    def test_final_optional_entity(self):
        """Ensure final optional entity is matched."""

        ini_text = """
        [ChangeDisplay]
        display (top | bottom){location} [(page | layer){layout}]
        """

        graph = intents_to_graph(parse_ini(ini_text))

        recognitions = zero_times(recognize("display bottom layer", graph))

        self.assertEqual(len(recognitions), 1)
        recognition = recognitions[0]
        self.assertIsNotNone(recognition.intent)

        entities = {e.entity: e for e in recognition.entities}

        self.assertIn("location", entities)
        location = entities["location"]
        self.assertEqual(location.value, "bottom")

        self.assertIn("layout", entities)
        layout = entities["layout"]
        self.assertEqual(layout.value, "layer")


# -----------------------------------------------------------------------------


def zero_times(recognitions):
    """Set times to zero so they can be easily compared in assertions"""
    for recognition in recognitions:
        recognition.recognize_seconds = 0

    return recognitions
